@@ -1,10 +1,11 @@
 package Log::Log4perl::Tiny;
-$Log::Log4perl::Tiny::VERSION = '1.2.8'; # TRIAL
+$Log::Log4perl::Tiny::VERSION = '1.3.0'; # TRIAL
 # ABSTRACT: mimic Log::Log4perl in one single module
 
 use warnings;
 use strict;
 use Carp;
+use POSIX ();
 
 our ($TRACE, $DEBUG, $INFO, $WARN, $ERROR, $FATAL, $OFF, $DEAD);
 my ($_instance, %name_of, %format_for, %id_for);
@@ -122,6 +123,11 @@ sub new {
 
    $self->format('[%d] [%5p] %m%n') unless exists $self->{format};
 
+   if (exists $args{loglocal}) {
+      my $local = $args{loglocal};
+      $self->loglocal($_, $local->{$_}) for keys %$local;
+   }
+
    return $self;
 } ## end sub new
 
@@ -188,6 +194,15 @@ sub LEVELNAME_FOR {
    return;
 }
 
+sub loglocal {
+   my $self = shift;
+   my $key = shift;
+   my $retval = delete $self->{loglocal}{$key};
+   $self->{loglocal}{$key} = shift if @_;
+   return $retval;
+}
+sub LOGLOCAL { return get_logger->loglocal(@_) }
+
 sub format {
    my $self = shift;
 
@@ -195,22 +210,50 @@ sub format {
       $self->{format} = shift;
       $self->{args} = \my @args;
       my $replace = sub {
-         my ($num, $op) = @_;
-         return '%%' if $op eq '%';
-         return "%%$op" unless defined $format_for{$op};
-         push @args, $op;
-         return "%$num$format_for{$op}[0]";
+         if (defined $_[2]) { # op with options
+            my ($num, $opts, $op) = @_[0..2];
+            push @args, [$op, $opts];
+            return "%$num$format_for{$op}[0]";
+         }
+         if (defined $_[4]) { # op without options
+            my ($num, $op) = @_[3,4];
+            push @args, [$op];
+            return "%$num$format_for{$op}[0]";
+         }
+
+         # not an op
+         my $char = ((! defined($_[5])) || ($_[5] eq '%')) ? '' : $_[5];
+         return '%%' . $char; # keep the percent AND the char, if any
       };
 
       # transform into real format
-      my $format_chars = join '', keys %format_for;
-      $self->{format} =~ s{
+      my ($with_options, $standalone) = ('', '');
+      for my $key (keys %format_for) {
+         my $type = $format_for{$key}[2] || '';
+         $with_options .= $key if $type;
+         $standalone   .= $key if $type ne 'required';
+      }
+      # quotemeta or land on impossible character class if empty
+      $_ = length($_) ? quotemeta($_) : '^\\w\\W'
+         for ($with_options, $standalone);
+      $self->{format} =~ s<
             %                      # format marker
-            ( -? \d* (?:\.\d+)? )  # number
-            ([$format_chars])      # specifier
-         }
+            (?:
+                  (?:                       # something with options
+                     ( -? \d* (?:\.\d+)? )  # number
+                     ( (?:\{ .*? \}) )      # options
+                     ([$with_options])     # specifier
+                  )
+               |  (?:
+                     ( -? \d* (?:\.\d+)? )  # number
+                     ([$standalone])        # specifier
+                  )
+               |  (.)                       # just any char
+               |  \z                        # just the end of it!
+            )
+         >
          {
-            $replace->($1, $2);
+            $replace->($1, $2, $3, $4, $5, $6);
          }gsmex;
    } ## end if (@_)
    return $self->{format};
@@ -238,9 +281,10 @@ sub log {
    my %data_for = (
       level   => $level,
       message => \@_,
+      (exists($self->{loglocal}) ? (loglocal => $self->{loglocal}) : ()),
    );
    my $message = sprintf $self->{format},
-     map { $format_for{$_}[1]->(\%data_for); } @{$self->{args}};
+     map { $format_for{$_->[0]}[1]->(\%data_for, @$_); } @{$self->{args}};
 
    return $self->emit_log($message);
 } ## end sub log
@@ -380,14 +424,16 @@ BEGIN {
 
    # Time tracking's start time. Used to be tied to $^T but Log::Log4perl
    # does differently and uses Time::HiRes if available
-   my $start_time = time(); # default, according to Log::Log4perl
    my $has_time_hires;
+   my $gtod = sub { return (time(), 0) };
    eval {
       require Time::HiRes;
       $has_time_hires = 1;
-      $start_time = [ Time::HiRes::gettimeofday() ];
+      $gtod = \&Time::HiRes::gettimeofday;
    };
 
+   my $start_time = [ $gtod->() ];
+
    # For supporting %R
    my $last_log = $start_time;
 
@@ -407,11 +453,38 @@ BEGIN {
       ],
       d => [
          s => sub {
-            my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday,
-               $isdst) = localtime();
-            sprintf '%04d/%02d/%02d %02d:%02d:%02d',
-              $year + 1900, $mon + 1, $mday, $hour, $min, $sec;
-           }
+            my ($epoch) = @{shift->{tod} ||= [$gtod->()]};
+            return POSIX::strftime('%Y/%m/%d %H:%M:%S', localtime($epoch));
+         },
+      ],
+      D => [
+         s => sub {
+            my ($data, $op, $options) = @_;
+            $options = '{}' unless defined $options;
+            $options = substr $options, 1, length($options) - 2;
+            my %flag_for = map {$_ => 1} split /\s*,\s*/, lc($options);
+            my ($s, $u) = @{$data->{tod} ||= [$gtod->()]};
+            $u = substr "000000$u", -6, 6; # padding left with 0
+            return POSIX::strftime("%Y-%m-%d %H:%M:%S.$u+0000", gmtime $s)
+               if $flag_for{utc};
+            return POSIX::strftime("%Y-%m-%d %H:%M:%S.$u%z", localtime $s)
+         },
+         'optional'
+      ],
+      e => [
+         s => sub {
+            my ($data, $op, $options) = @_;
+            $data->{tod} ||= [$gtod->()]; # guarantee consistency here
+            my $local = $data->{loglocal} or return '';
+            my $key = substr $options, 1, length($options) - 2;
+            return '' unless exists $local->{$key};
+            my $target = $local->{$key};
+            return '' unless defined $target;
+            my $reft = ref $target or return $target;
+            return '' unless $reft eq 'CODE';
+            return $target->($data, $op, $options);
+         },
+         'required',
       ],
       F => [
          s => sub {
@@ -470,32 +543,25 @@ BEGIN {
       n => [s => sub { "\n" },],
       p => [s => sub { $name_of{shift->{level}} },],
       P => [d => sub { $$ },],
-      r => [d => ( $has_time_hires # install sub depending on Time::HiRes
-         ?  sub {
-               my ($s, $m) = Time::HiRes::gettimeofday();
-               $s -= $start_time->[0];
-               $m = int(($m - $start_time->[1]) / 1000);
-               ($s, $m) = ($s - 1, $m + 1000) if $m < 0;
-               return $m + 1000 * $s;
-            }
-         :  sub {
-               return 1000 * (time() - $start_time);
-            }
-      ) ],
-      R => [d => ( $has_time_hires # install sub depending on Time::HiRes
-         ?  sub {
-               my ($sx, $mx) = Time::HiRes::gettimeofday();
-               my $s = $sx - $last_log->[0];
-               my $m = int(($mx - $last_log->[1]) / 1000);
-               ($s, $m) = ($s - 1, $m + 1000) if $m < 0;
-               $last_log = [ $sx, $mx ];
-               return $m + 1000 * $s;
-            }
-         :  sub {
-               my $l = $last_log;
-               return 1000 * (($last_log = time()) - $l);
-            }
-      ) ],
+      r => [
+         d => sub {
+            my ($s, $u) = @{shift->{tod} ||= [$gtod->()]};
+            $s -= $start_time->[0];
+            my $m = int(($u - $start_time->[1]) / 1000);
+            ($s, $m) = ($s - 1, $m + 1000) if $m < 0;
+            return $m + 1000 * $s;
+         },
+      ],
+      R => [
+         d => sub {
+            my ($sx, $ux) = @{shift->{tod} ||= [$gtod->()]};
+            my $s = $sx - $last_log->[0];
+            my $m = int(($ux - $last_log->[1]) / 1000);
+            ($s, $m) = ($s - 1, $m + 1000) if $m < 0;
+            $last_log = [ $sx, $ux ];
+            return $m + 1000 * $s;
+         },
+      ],
       T => [
          s => sub {
             my $level = 4;
@@ -575,7 +641,7 @@ Log::Log4perl::Tiny - mimic Log::Log4perl in one single module
 
 =head1 VERSION
 
-version 1.2.8
+version 1.3.0
 
 =head1 SYNOPSIS
 
@@ -634,6 +700,20 @@ version 1.2.8
    my @queue; # e.g. all log messages will be put here
    $logger->fh(sub { push @queue, $_[0] });
 
+   # As of 1.4.0, you can set key-value pairs in the logger
+   $logger->loglocal(foo => 'bar');
+   LOGLOCAL(baz => 100);
+
+   # You can later retrieve the value in the format with %{key}e
+   $logger->format("[%{foo}e] [%{baz}e] %m%n");
+
+   # You are not limited to scalars, you can use references too
+   LOGLOCAL(baz => sub {
+      my ($data, $op, $ekey) = @_;
+      return join '.', @{$data->{tod}}; # epoch from gettimeofday
+   });
+   LOGLOCAL(foo => sub { return rand 100 });
+
 =head1 DESCRIPTION
 
 Yes... yet another logging module. Nothing particularly fancy nor
@@ -934,6 +1014,16 @@ C<$Log::Log4perl::LOGEXIT_CODE>, or set a code with
 C<logexit_code()> - but you have to wait to read something about the
 object-oriented interface before doing this!
 
+As indicated, functions L</LOGWARN>, L</LOGDIE>, L</LOGCARP>,
+L</LOGCLUCK>, L</LOGCROACK>, and L</LOGCONFESS> (as well as their
+lowercase counterparts called as object methods) both emit the log
+message on the normal output channel for Log::Log4perl::Tiny and call
+the respective function. This might not be what you want in the default
+case where the output channel is standard error, because you will end up
+with duplicate error messages. You can avoid the call to the
+I<canonical> function setting import option C<:no_extra_logdie_message>,
+in line with what L<Log::Log4perl> provides.
+
 There is also one additional stealth function that L<Log::Log4perl>
 misses but that I think is of the outmoste importance: C<LOGLEVEL>, to
 set the log level threshold for printing. If you want to be 100%
@@ -972,6 +1062,9 @@ that are modeled (with simplifications) after L<Log::Log4perl>'s ones:
     %c Category of the logging event.
     %C Fully qualified package (or class) name of the caller
     %d Current date in yyyy/MM/dd hh:mm:ss format
+    %D Current date in 
+    %{type}D Current date as strftime's "%Y-%m-%d %H:%M:%S.$u%z"
+    %{key}e Evaluate or substitute (extension WRT Log::Log4perl)
     %F File where the logging event occurred
     %H Hostname
     %l Fully qualified name of the calling method followed by the
@@ -989,13 +1082,57 @@ that are modeled (with simplifications) after L<Log::Log4perl>'s ones:
        a %R to current logging event
     %% A literal percent (%) sign
 
-Notably, both C<%x> (NDC) and C<%X> (MDC) are missing. Moreover, the
-extended specifier feature with additional info in braces (like
-C<%d{HH:mm}>) is missing, i.e. the structure of each specifier above
-is fixed. (Thanks to C<Log::Tiny> for the cool trick of how to handle
-the C<printf>-like string, which is probably mutuated from
+Notably, both C<%x> (NDC) and C<%X> (MDC) are missing. The functionality
+for the latter is partially covered by the extension C<%e> explained
+below.  Moreover, the extended specifier feature with additional info in
+braces (like C<%d{HH:mm}>) is missing, i.e. the structure of each
+specifier above is fixed. (Thanks to C<Log::Tiny> for the cool trick of
+how to handle the C<printf>-like string, which is probably mutuated from
 C<Log::Log4perl> itself according to the comments).
 
+There are also two extensions with respect to Log::Log4perl, that help
+partially cover the missing items explained above, as of release 1.4.0:
+
+=over
+
+=item C<%D>
+
+=item C<%{type}D>
+
+expanded to a timestamp according to L<POSIX/strftime> specifier
+C<%Y-%m-%d %H:%M:%S.$u%z>, i.e. a timestamp that includes up to the
+microsecond (on platform where this is available, otherwise zeros will
+be used for sub-second values). By default the local time is used, but
+you can also pass a C<type> specifier set to the string C<utc>, in which
+case the UTC time will be used (via C<gmtime>).
+
+=item C<%{key}e>
+
+expanded according to what set via L</loglocal>/L</LOGLOCAL>. These two
+functions allow setting key-value pairs; the C<key> is used to find the
+associated value, then the value is returned as-is if it's a simple
+defined scalar, otherwise if it is a sub reference it is invoked,
+otherwise the empty string is returned.
+
+In case a subroutine reference is set, it is called with the following
+parameters:
+
+   $sub->($data, $op, $options);
+
+where C<$data> is a reference to a hash that contains at least the
+C<tod> key, associated to an array with the output of C<gettimeofday>
+(if L<Time::HiRes> is available) or its equivalent (if L<Time::HiRes> is
+not available), C<$op> is the letter C<e> and C<$options> is the string
+containing the C<key> in braces (e.g. C<{this-is-the-key}>).
+
+=back
+
+As of release 1.4.0 all time-expansions in a single log refer to the
+same time, i.e. if you specify the format string C<%D %D> and you have
+microsecond-level resolution, the two values in output will be the same
+(as opposed to show two slightly different times, related to the
+different expansion times of the C<%D> specifier).
+
 =head1 INTERFACE 
 
 You have two interfaces at your disposal, the functional one (with all
@@ -1076,6 +1213,13 @@ emit log at C<FATAL> level and then call C<Carp::confess()>;
 
 set the minimum log level for sending a log message to the output;
 
+=item C<< LOGLOCAL >>
+
+(Not in L<Log::Log4perl>) (Not imported with C<:easy>) (As of 1.4.0) 
+
+set a key-value pair useful for later expansion via code C<%{key}e>. See
+L</loglocal> below;
+
 =item C<< build_channels >>
 
 (Not in L<Log::Log4perl>) (Not imported with C<:easy>)
@@ -1174,6 +1318,10 @@ following keys:
 set a list (through an array reference) of channels. See
 L</build_channels> for additional information.
 
+=item B<< fh >>
+
+see method C<fh> below
+
 =item B<< file >>
 
 =item B<< file_insecure >>
@@ -1200,9 +1348,10 @@ precedence over C<fh> described below.
 see L<< C<easy_init()> >> and the methods below with the same
 name
 
-=item B<< fh >>
+=item B<< loglocal >>
 
-see method C<fh> below
+pass a reference to a hash with key-value pairs to be set via
+L</loglocal>;
 
 =back
 
@@ -1352,6 +1501,23 @@ get/set the line formatting;
 get/set the exit code to be used with C<logexit()> (and
 C<logdie()> as well if C<die()> doesn't exit).
 
+=item C<< loglocal >>
+
+get/set a local key-value pair for expansion with C<%{key}e>.
+
+Always returns the previous value associated to the provided key,
+removing it:
+
+   my $value = $logger->loglocal('some-key');
+   # now, 'some-key' does not exist any more and has no value associated
+
+If you pass a value too, it will be set:
+
+   $logger->loglocal(foo => 'bar');
+   my $old = $logger->loglocal(foo => 'whatever');
+   # $old is 'bar'
+   # current value associated to foo is 'whatever'
+
 =back
 
 =head1 DEPENDENCIES
